use anyhow::{Context, Result, bail};
use cache::relative_to_absolute;
use clap::{Parser, Subcommand};
use std::collections::HashMap;
use std::env;
use std::fs::File;
use std::io::{BufRead, BufReader};
use std::path::PathBuf;
use std::process::Command;

mod cache;

// I'm not sure if reqwest::blocking's write_to is working right or not.
// It seems like it might be storing the entire file in memory...
// Which I don't want.  I might have to go back to tokio and chunk.

// see reqwest/web-o/src/cache.rs for example cache
// It restores reqwest::header::HeaderMap
// (which allows duplicates... and ignores case on keys)

#[derive(Parser)]
#[command(
    about = "Go Updater",
    long_about = "Go Updater

This checks the https://go.dev/dl for newer versions of go.
It depends upon the GOPATH and GOROOT environment variables being set.

This can't update a package manager installed version of go (permissions)."
)]
struct Cli {
    /// Cache directory path
    #[arg(short, long, default_value = "cache")]
    cache: PathBuf,

    // If I make this optional, I can't seem to get the display_help code
    // to work now.  Not sure why.
    #[command(subcommand)]
    // command: Commands,
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    /// Update go
    Update {},
    /// Display information
    Info {},
}

/// Query `go env`
#[must_use]
fn get_go_env() -> Result<HashMap<String, String>> {
    let output = Command::new("go").arg("env").output()?;
    if output.status.success() {
        let mut result = HashMap::<String, String>::new();
        for mut line in String::from_utf8_lossy(output.stdout.as_slice()).split("\n") {
            line = line.trim();
            if let Some(parts) = line.split_once("=") {
                let value = parts.1.trim_matches('\'');
                result.insert(parts.0.to_string(), value.to_string());
            }
        }
        Ok(result)
    } else {
        bail!("Failed to query `go env`");
    }
}

/// Query `go version`
#[must_use]
#[allow(dead_code)]
#[deprecated = "Use get_go_env"]
fn find_go_version() -> Result<String> {
    let output = Command::new("go").arg("version").output()?;
    if output.status.success() {
        // Ok! We have something!
        return Ok(String::from_utf8_lossy(output.stdout.as_slice()).to_string());
    }
    bail!("Failed to query go version.");
}

/// Locate the go binary
///
/// This is redundant, it should be located via GO_PATH.
#[allow(dead_code)]
#[must_use]
fn find_go() -> Result<String> {
    let output = Command::new("which").arg("go").output()?;
    if output.status.success() {
        return Ok(String::from_utf8_lossy(output.stdout.as_slice()).to_string());
    }
    bail!("Failed to locate go binary.");
}

const GO_URL: &str = "https://go.dev/dl/";
// const GO_FILE: &str = "go-dl.html";

// 2 MB download of html...

// static APP_USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"),);

/*
CSS Selector elements of interest:  (a href="" part).
href are relative...

div[class="downloadWrapper"] > a[class="download downloadBox"]

<a class="download downloadBox" href="/dl/go1.24.0.windows-amd64.msi">
<a class="download downloadBox" href="/dl/go1.24.0.linux-amd64.tar.gz">

table[class="downloadtable"] > tr[class=" "] > td[class="filename"] > a[class="download"]

Or possibly,

table[class="downloadtable"] > tr[class=" "]
or tr[class="highlight "] ?
and grab the section of td's.  class=filename has a href, last has SHA256.

<a class="download" href="/dl/go1.24.0.src.tar.gz">go1.24.0.src.tar.gz</a>
<a class="download" href="/dl/go1.24.0.linux-amd64.tar.gz">go1.24.0.linux-amd64.tar.gz</a>
<a class="download" href="/dl/go1.24.0.linux-arm64.tar.gz">go1.24.0.linux-arm64.tar.gz</a>
<a class="download" href="/dl/go1.24.0.windows-amd64.zip">go1.24.0.windows-amd64.zip</a>
<a class="download" href="/dl/go1.24.0.windows-amd64.msi">go1.24.0.windows-amd64.msi</a>

 */

/*
fn download_and_save(url: &str, filename: &str) -> Result<()> {
    let client = reqwest::blocking::Client::builder()
        .user_agent(APP_USER_AGENT)
        .build()?;
    print!("Downloading: {url} ");
    let _ = stdout().flush();

    let mut resp = client.get(url).send().context("Failed get")?;

    if resp.status().is_success() {
        let mut file =
            File::create(filename).with_context(|| format!("Creating file {filename} failed."))?;
        resp.copy_to(&mut file)?;
    } else {
        bail!("Status Code: {:?}", resp.status());
    }
    println!("OK");

    Ok(())
}
*/

// URL: https://go.dev/dl/go1.24.1.linux-amd64.tar.gz
#[must_use]
/// Get go version from download URL.
fn version_from_url(url: &str, arch: &str) -> Option<String> {
    if let Some(parts) = url.split_once(arch) {
        if let Some(part) = parts.0.rsplit_once("/go") {
            let part = part.1.trim_matches('.');
            return Some(part.to_string());
        }
    }
    None
}

#[must_use]
/// Get go version from `go version` output.
fn version_from_go(text: &str) -> Option<String> {
    let parts: Vec<&str> = text.split(' ').collect();
    if parts.len() == 4 {
        return Some(parts[2].to_string().replace("go", ""));
    }
    None
}

/// Return just the href="<return this part>".
#[must_use]
fn just_href(link: &str) -> Result<String> {
    let parts = link.split_once("href=\"").unwrap_or(("", "")).1;
    let href = parts.split_once("\"").unwrap_or(("", "")).0;
    if !href.is_empty() {
        return Ok(href.to_string());
    }
    bail!("Unable to locate href");
}

/// Find a href link for given arch (architecture)
///
/// Look for <a class="download" href="
#[must_use]
fn find_arch_link(arch: &str, fp: &File) -> Result<String> {
    let reader = BufReader::new(fp);
    for line in reader.lines() {
        if let Ok(line) = line {
            if line.contains("a class=\"download\"") {
                if line.contains(arch) {
                    // Return just the href part.
                    return just_href(&line);
                }
            }
        }
    }
    bail!("Unable to locate architecture download link");
}

/*
/// find_link for given arch (architecture)
///
/// Look for <a class="download" href=""
#[must_use]
#[deprecated]
fn find_link(arch: &str) -> Result<String> {
    // , Box<dyn Error>> {
    let fp = File::open(GO_FILE)?;
    let reader = BufReader::new(fp);
    for line in reader.lines() {
        if let Ok(line) = line {
            if line.contains("a class=\"download\"") {
                if line.contains(arch) {
                    // Return just the href part.
                    return just_href(&line);
                }
            }
        }
    }
    bail!("Unable to locate architecture download link");
}
*/

/// Get value from given HashMap<String,String>
/// 
/// This gives us a String, not a &String, and not an Option.
fn hashget(hash: &HashMap<String,String>, key: &str) -> String {
    if let Some(value) = hash.get(key) {
        return value.clone();
    }
    String::new()
}

fn main() -> Result<()> {
    let cli = Cli::parse();

    /*
    GOROOT, defaults to $HOME/go
    This seems to be managed by go now.  So running ```go env``` would
    have/show GOROOT being set by that instance of go.
    GOPATH is another story.  It's where go stores things it downloads from
    the web, and most certainly shouldn't be $HOME/go (!= GOROOT).

    I want to be a bit looser on requiring GOROOT later on.
    */

    // Get go environment
    let go_path = env::var("GOPATH").unwrap_or(String::new());
    let go_root = env::var("GOROOT").unwrap_or(String::new());

    let go_env = get_go_env()
        .context("Calling `go env` failed.")?;
    let mut go_version = hashget(&go_env, "GOVERSION");
    if go_version.starts_with("go") {
        go_version.remove(0);
        go_version.remove(0);
    }

    let go_arch = hashget(&go_env, "GOARCH");
    let go_os = hashget(&go_env, "GOOS");
    let go_path = hashget(&go_env, "GOPATH");
    let go_root = hashget(&go_env, "GOROOT");
    let go_os_arch = format!("{go_os}-{go_arch}");

    println!("{} - {}-{}", go_version, go_os, go_arch);

    // Initialize the cache.
    let cache = cache::Cache::new(cli.cache, None)?;

    match &cli.command {
        Some(Commands::Update {}) => {
            let status = cache.fetch(GO_URL);

            // Check to see if file already exists AND check version against go's version.
            // Since the go.dev site doesn't allow caching or knowing when it changes,
            // we always end up in Status::Fetched arm.

            match status {
                cache::Status::Fetched(filename) => {
                    let fp = File::open(filename)?;
                    let link = find_arch_link(&go_os_arch, &fp);
                    if let Ok(relative) = link {
                        // Download link for arch located.  Make useable.
                        let abs = relative_to_absolute(GO_URL, &relative).unwrap();
                        println!("URL: {}", abs);

                        let latest_version = version_from_url(&abs, &go_os_arch);
                        if let Some(latest) = latest_version {
                            println!("Version: {} [have {}]", latest, go_version);
                            if go_version != latest {
                                println!("Downloading newer version...");
                                let latest_status = cache.fetch(&abs);
                                println!("Latest: {:?}", latest_status);
                            }
                        } else {
                            println!("Finding version failed for string: [{}]", abs);
                        }
                    } else {
                        bail!("Unable to locate download link");
                    }
                }
                cache::Status::Cached(filename) => {
                    println!("(from cache)"); // I wish I could see this.
                    let fp = File::open(filename)?;
                    let link = find_arch_link(&go_os_arch, &fp);
                    if let Ok(relative) = link {
                        let abs = relative_to_absolute(GO_URL, &relative).unwrap();
                        println!("URL: {}", abs);
                    }
                }
                _ => {
                    println!("Status = {:?}", status);
                }
            }
        }
        Some(Commands::Info {}) => {
            println!("GOPATH      {}", go_path);
            println!("GOROOT      {}", go_root);
            println!("go ver      {}", go_version);
            println!("go_arch     {}", go_arch);
            println!("go os       {}", go_os);
            println!("go os arch: {}", go_os_arch);
        }
        None => {
            // Display help.
            let _show_help: Cli = Cli::parse_from(["", "--help"]);
        }
    }
    Ok(())
}
